"""
    @Title:             Database creation + handling of patients who failed.

    @Description:       Create the (possibly partial) HDF5 database, update the tag values dictionary using the
                        information from the failed patients' images, and finally create the full HDF5 database.
"""

import env_examples  # Modifies path, DO NOT REMOVE

import json
from typing import Dict, List, Union

from delia.databases import PatientsDatabase
from delia.extractors import PatientsDataExtractor, PatientWhoFailed


def get_updated_tag_values(
        patients_who_failed: List[PatientWhoFailed],
        tag_values: Union[str, Dict[str, List[str]]]
) -> Union[str, Dict[str, List[str]]]:
    """
    Add a tag value to the tag value list of the given series key.

    Parameters
    ----------
    patients_who_failed : List[PatientWhoFailed]
        List of patients with one or more images not added to the HDF5 database due to the absence of the series in
        the patient record.
    tag_values : Union[str, Dict[str, List[str]]], default = None.
        A dictionary that contains the desired tag's values for the images that absolutely needs to be extracted
        from the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
        values associated with the specified tag. The images associated with these tag values do not need to have a
        corresponding segmentation. In fact, the whole point of adding a way to specify the values for a specified tag
        that must be added to the database is to be able to add images without their segmentation.
    """
    for patient in patients_who_failed:
        for image, image_tag_values in patient.failed_images.items():
            if not any([series in tag_values[image] for series in patient.available_tag_values]):
                while True:
                    new_tag_value = input(
                        f"\nNo available series for {image}. \nAvailable series are "
                        f"{patient.available_tag_values}. \nName of the tag value to add : ")

                    print(f"Given tag value name is {new_tag_value}.")

                    if new_tag_value in patient.available_tag_values:
                        print(f"tag value successfully added to the tag values json file.")
                        break
                    else:
                        print(f"Invalid tag value! \n{new_tag_value} not found in the patient's "
                              f"dicom files. Please try again.")

                tag_values[image] += [new_tag_value]

    return tag_values


if __name__ == "__main__":
    # ----------------------------------------------------------------------------------------------------------- #
    #     Create database (some images of some patients might fail to be added to the database due to the         #
    #                               absence of the tag value in the patient record)                               #
    # ----------------------------------------------------------------------------------------------------------- #
    patients_data_extractor = PatientsDataExtractor(
        path_to_patients_folder="data/patients",
        tag="SeriesDescription",
        tag_values="data/incorrect_tag_values.json"
    )

    database = PatientsDatabase(path_to_database="data/patients_database.h5")

    patients_who_failed = database.create(
        patients_data_extractor=patients_data_extractor,
        overwrite_database=True
    )

    # Print list of patients who failed
    print(f"Patients who failed the pipeline : {patients_who_failed}")

    # ----------------------------------------------------------------------------------------------------------- #
    #                                              Update tag values                                              #
    # ----------------------------------------------------------------------------------------------------------- #
    with open("data/incorrect_tag_values.json", "r") as json_file:
        tag_values = json.load(json_file)

    updated_tag_values = get_updated_tag_values(
        patients_who_failed=patients_who_failed,
        tag_values=tag_values
    )

    with open("data/incorrect_tag_values.json", 'w', encoding='utf-8') as json_file:
        json.dump(updated_tag_values, json_file, ensure_ascii=False, indent=4)

    # ----------------------------------------------------------------------------------------------------------- #
    #                                           Create complete database                                          #
    # ----------------------------------------------------------------------------------------------------------- #
    patients_data_extractor = PatientsDataExtractor(
        path_to_patients_folder="data/patients",
        tag_values=updated_tag_values
    )

    patients_who_failed = database.create(
        patients_data_extractor=patients_data_extractor,
        tags_to_use_as_attributes=[(0x0008, 0x103E), (0x0020, 0x000E), (0x0008, 0x0060)],
        overwrite_database=True
    )

    # Print list of patients who failed. This list should be empty now.
    print(f"Patients who failed the pipeline : {patients_who_failed}")



"""
    @file:              dicom_reader.py
    @Author:            Maxence Larose

    @Creation Date:     10/2021
    @Last modification: 03/2022

    @Description:       This file contains the DicomReader class which is used to read dicom files contained in a given
                        folder.
"""

from collections import defaultdict
from glob import glob
import logging
import os
from typing import Dict, List, NamedTuple, Optional, Set, Tuple, Union

import pydicom
import SimpleITK as sitk

from delia.utils.data_model import ImageDataModel
from delia.readers.segmentation.segmentation_strategy import SegmentationStrategies
from delia.utils.tools import is_path_valid

_logger = logging.getLogger(__name__)


class DicomReader:
    """
    A class used to read dicom files contained in a given folder.
    """

    UNKNOWN = "Unknown"

    class SeriesData(NamedTuple):
        """
        Tag value namedtuple to simplify management of values.
        """
        tag_value: str
        paths_to_dicoms_from_series: List[str]
        dicom_header: pydicom.dataset.FileDataset

    def __init__(
            self,
            path_to_patient_folder: str,
            tag: Union[str, Tuple[int, int]],
            load_segmentations: bool = True
    ):
        """
        Constructor of the class DicomReader.

        Parameters
        ----------
        path_to_patient_folder : str
            Path to the folder containing the patient DICOM files.
        tag : Union[str, Tuple[int, int]]
            Keyword or tuple of the DICOM tag to use while selecting which files to extract.
        load_segmentations : bool
            Whether to load segmentations or not.
        """
        super(DicomReader, self).__init__()

        is_path_valid(path=path_to_patient_folder)
        self._path_to_patient_folder = path_to_patient_folder
        self._load_segmentations = load_segmentations
        self.tag = tag

        self._images_series_data_dict = {}
        self._segmentations_series_data_dict = {}
        self.__series_data_dict = self.__get_series_data_dict()

    @staticmethod
    def get_dicom_header(
            path_to_dicom: str,
            show: bool = False
    ) -> pydicom.dataset.FileDataset:
        """
        Get a dicom header given the path to the dicom.

        Parameters
        ----------
        path_to_dicom : str
            The path to the dicom file.
        show : bool
            Show dicom header in console.

        Returns
        -------
        loaded_dicom : pydicom.dataset.FileDataset
            Loaded DICOM dataset.
        """
        loaded_dicom = pydicom.dcmread(path_to_dicom, stop_before_pixels=True)

        if show:
            _logger.info(loaded_dicom)

        return loaded_dicom

    @staticmethod
    def _get_tag_value(dicom_header: pydicom.FileDataset, tag: Union[str, Tuple[int, int]]) -> str:
        """
        Get tag value of given DICOM header.

        Parameters
        ----------
        dicom_header : pydicom.dataset.FileDataset
            Loaded DICOM dataset.
        tag : Union[str, Tuple[int, int]]
            Keyword or tuple of the DICOM tag to use while selecting which files to extract.

        Returns
        -------
        tag_value : str
            tag value.
        """
        if hasattr(dicom_header, tag):
            if isinstance(dicom_header[tag].value, str):
                return dicom_header[tag].value
            else:
                return dicom_header[tag].repval
        else:
            return DicomReader.UNKNOWN

    def __get_paths_to_dicoms_from_series(self, path_to_dicom_folder: str) -> Dict[str, List[str]]:
        """
        Get all series IDs from a patient's dicom folder.

        Parameters
        ----------
        path_to_dicom_folder : str
            Path to a folder containing only DICOM files.

        Returns
        -------
        paths_to_dicoms_from_series : Dict[str, List[str]]
            Dictionary of lists of paths to dicoms from each series.
        """
        paths = [
            y for x in os.walk(path_to_dicom_folder) for y in glob(os.path.join(x[0], '*')) if os.path.isfile(y) is True
        ]

        paths_and_headers_for_each_series = defaultdict(list)
        for path in paths:
            dicom_header = self.get_dicom_header(path_to_dicom=path)
            paths_and_headers_for_each_series[dicom_header.SeriesInstanceUID].append((path, dicom_header))

        paths_to_dicoms_from_series = defaultdict(list)
        for series_id, result in paths_and_headers_for_each_series.items():
            if len(result) != 1:
                if hasattr(result[1][1], "SliceLocation"):
                    paths_to_dicoms_from_series[series_id] = [
                        i[0] for i in sorted(result, key=lambda s: s[1].SliceLocation)
                    ]
                else:
                    paths_to_dicoms_from_series[series_id] = [
                        i[0] for i in result
                    ]
            else:
                paths_to_dicoms_from_series[series_id] = [result[0][0]]

        return paths_to_dicoms_from_series

    def __get_series_data_dict(self) -> Dict[str, SeriesData]:
        """
        Get the series data from series IDs and the path to the patient's dicom folders.

        Returns
        -------
        series_data_dict : Dict[str, SeriesData]
            Dictionary of the data from the selected series.
        """
        series_data_dict: Dict[str, DicomReader.SeriesData] = {}
        all_patient_ids: Set[str] = set()
        for series_id, paths in self.__get_paths_to_dicoms_from_series(self._path_to_patient_folder).items():
            path_to_first_dicom_of_series = paths[0]
            loaded_dicom_header = self.get_dicom_header(path_to_dicom=path_to_first_dicom_of_series)
            all_patient_ids.add(loaded_dicom_header.PatientID)

            series_data = self.SeriesData(
                tag_value=self._get_tag_value(loaded_dicom_header, self.tag),
                paths_to_dicoms_from_series=paths,
                dicom_header=loaded_dicom_header
            )
            series_data_dict[series_id] = series_data

        if len(all_patient_ids) == 0:
            raise AssertionError(f"The patient folder must contain DICOM files. This is not the case for the patient "
                                 f"whose data is currently being downloaded since no DICOM files are found in its "
                                 f"folder.")
        elif len(all_patient_ids) > 1:
            raise AssertionError(f"All DICOM files in the same folder must belong to the same patient. This is not "
                                 f"the case for the patient whose data is currently being downloaded since the "
                                 f"identifiers {all_patient_ids} are found in their folder.")

        for series_id, series_data in series_data_dict.items():
            if series_data.dicom_header.Modality in SegmentationStrategies.get_available_modalities():
                if self._load_segmentations:
                    self._segmentations_series_data_dict[series_id] = series_data
            else:
                self._images_series_data_dict[series_id] = series_data

        return series_data_dict

    @staticmethod
    def __get_3d_sitk_image_from_dicom_series(
            paths_to_dicoms_from_series: List[str],
    ) -> sitk.Image:
        """
        Get a 3D image array from a list of dicom paths associated to the same series.

        Parameters
        ----------
        paths_to_dicoms_from_series : List[str]
            A list of the paths to each dicom contained in the series.

        Returns
        -------
        image : sitk.Image
            3D SimpleITK image obtained from the series.
        """
        series_reader = sitk.ImageSeriesReader()
        series_reader.SetFileNames(paths_to_dicoms_from_series)

        image = series_reader.Execute()

        return image

    def get_dicom_headers(self, remove_segmentations: bool = False) -> List[pydicom.dataset.FileDataset]:
        """
        List of all the patient's dicom headers.

        Parameters
        ----------
        remove_segmentations : bool
            Whether or not to keep the segmentation files headers, default = False.

        Returns
        -------
        dicom_headers : List[pydicom.dataset.FileDataset]
            A list of the patient's DICOM headers.
        """
        dicom_headers = []
        for idx, series_data in enumerate(self.__series_data_dict.values()):
            if idx == 0:
                _logger.info(f"Patient ID : {series_data.dicom_header.PatientID}")
                _logger.debug(f"Path to images folder : {self._path_to_patient_folder}")

            _logger.debug(f"  {self.tag} : {series_data.tag_value}")

            if not remove_segmentations:
                dicom_headers.append(series_data.dicom_header)
            elif remove_segmentations:
                if series_data.dicom_header.Modality not in SegmentationStrategies.get_available_modalities():
                    dicom_headers.append(series_data.dicom_header)

        return dicom_headers

    def get_images_data(
            self,
            remove_segmentations: bool = False
    ) -> List[ImageDataModel]:
        """
        List of tuples containing simpleITK 3D images array and their corresponding dicom header. Each element in the
        list corresponds to an image series.

        Parameters
        ----------
        remove_segmentations : bool
            Whether to keep the segmentation files headers, default = False.

        Returns
        -------
        image_data_list : List[ImageDataModel]
            A list of the patient's images data.
        """
        images_data = []
        for idx, series_data in enumerate(self.__series_data_dict.values()):
            if idx == 0:
                _logger.info(f"Tag values found in the patient's images folder :")
            _logger.info(f"  {self.tag} {idx + 1}: {series_data.tag_value}")

            if remove_segmentations and (series_data.dicom_header.Modality in
                                         SegmentationStrategies.get_available_modalities()):
                pass
            elif series_data.tag_value == DicomReader.UNKNOWN:
                pass
            else:
                try:
                    image = self.__get_3d_sitk_image_from_dicom_series(
                        paths_to_dicoms_from_series=series_data.paths_to_dicoms_from_series
                    )

                    image_data = ImageDataModel(
                        dicom_header=series_data.dicom_header,
                        simple_itk_image=image,
                        paths_to_dicoms=series_data.paths_to_dicoms_from_series
                    )

                    images_data.append(image_data)

                except RuntimeError as e:
                    _logger.error(
                        f"      RuntimeError : {e}. Simple ITK raised an error while loading the series named "
                        f"{series_data.tag_value}. This series is therefore ignored."
                    )

        return images_data



"""
    @file:              patient_data_reader.py
    @Author:            Maxence Larose

    @Creation Date:     10/2021
    @Last modification: 10/2022

    @Description:       This file contains the PatientDataReader class which is used to read dicom files AND
                        segmentation files.
"""

import logging
from typing import Dict, List, Optional, Tuple, Union

from monai.transforms import Compose
from monai.transforms import MapTransform as MonaiMapTransform

from delia.readers.image.dicom_reader import DicomReader
from delia.readers.patient_data.patient_data_query_context import PatientDataQueryContext
from delia.transforms.applications import apply_transforms
from delia.transforms.data.transform import DataTransform
from delia.transforms.physical_space.transform import PhysicalSpaceTransform
from delia.utils.data_model import PatientDataModel
from delia.utils.transforms_history import TransformsHistory

_logger = logging.getLogger(__name__)


class PatientDataReader(DicomReader):
    """
    A class used to read dicom files AND segmentation files.
    """

    def __init__(
            self,
            path_to_patient_folder: str,
            tag_values: Optional[Dict[str, List[str]]],
            tag: Union[str, Tuple[int, int]],
            load_segmentations: bool = True,
            organs: Optional[List[str]] = None,
            erase_unused_dicom_files: bool = False
    ):
        """
        Used to check availability of given series' uid and tag values in the patient's dicom files.

        Parameters
        ----------
        path_to_patient_folder : str
            Path to the folder containing the patient's files.
        tag_values : Dict[str, List[str]]
            A dictionary that contains the desired tag's values for the images that absolutely needs to be extracted
            from the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            values associated with the specified tag.
        tag : Union[str, Tuple[int, int]]
            Keyword or tuple of the DICOM tag to use while selecting which files to extract.
        load_segmentations : bool = True
            Whether to load the segmentation files or not.
        organs : Optional[List[str]] = None
            List of organs to load. If None, all organs will be loaded.
        erase_unused_dicom_files: bool = False
            Whether to delete unused DICOM files or not. Use with caution.
        """
        super().__init__(
            path_to_patient_folder=path_to_patient_folder,
            tag=tag,
            load_segmentations=load_segmentations
        )

        self._images_dicom_headers = self.get_dicom_headers(remove_segmentations=True)
        self._tag_values = tag_values
        self._erase_unused_dicom_files = erase_unused_dicom_files
        self._organs = organs

        self.failed_images = []
        if tag_values is not None:
            self.check_availability_of_given_tag_value()

    @property
    def patient_id(self) -> str:
        """
        Patient ID.

        Returns
        -------
        patient_id : str
            Patient id.
        """
        return str(self._images_dicom_headers[0].PatientID)

    @property
    def paths_to_segmentations(self) -> List[str]:
        """
        Paths to segmentations.

        Returns
        -------
        paths_to_segmentations : List[str]
            List of paths to the segmentation files.
        """
        paths_to_segmentations = [
            series.paths_to_dicoms_from_series[0] for series in self._segmentations_series_data_dict.values()
        ]

        return paths_to_segmentations

    @property
    def tag_values(self) -> Dict[str, List[str]]:
        """
        Tag values setter.

        Returns
        -------
        tag_values : Dict[str, List[str]]
            A dictionary that contains the desired tag's values for the images that absolutely needs to be extracted
            from the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            values associated with the specified tag.
        """
        return self._tag_values

    @tag_values.setter
    def tag_values(self, tag_values: Dict[str, List[str]]):
        """
        Tag values setter.

        Parameters
        ----------
        tag_values : Dict[str, List[str]]
            A dictionary that contains the desired tag's values for the images that absolutely needs to be extracted
            from the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            values associated with the specified tag.
        """
        items = list(tag_values.items())
        for previous_items, current_items in zip(items, items[1:]):
            set_intersection = set(previous_items[1]) & set(current_items[1])

            if bool(set_intersection):
                raise AssertionError(
                    f"\nThe dictionary of tag values should not contain the same series names for different "
                    f"images/modalities. \nHowever, here we find the series names {previous_items[1]} for the "
                    f"{previous_items[0]} image and {current_items[1]} for the {current_items[0]} image. \nClearly, "
                    f"the images series values are overlapping because of the series named {set_intersection}."
                )

        self.tag_values = tag_values

    @property
    def available_tag_values(self) -> List[str]:
        """
        Available values of the specified tag.

        Returns
        -------
        available_tag_values : List[str]
            Available values of the specified tag in the patient dicom files.
        """
        available_tag_values = [
            self._get_tag_value(dicom_header, self.tag) for dicom_header in self._images_dicom_headers
        ]

        return available_tag_values

    def check_availability_of_given_tag_value(self) -> None:
        """
        Check availability of given value for specified tag in the patient's dicom files.
        """
        _logger.debug("Checking availability of given value for specified tag...")
        for series_key, tag_value_list in self.tag_values.items():
            if any(series in self.available_tag_values for series in tag_value_list):
                pass
            else:
                self.record_failed_images(series_key)
        _logger.debug("Done.")

    def record_failed_images(self, series_key: str) -> None:
        """
        Record failed images.

        Parameters
        ----------
        series_key : str
            Series key.
        """
        _logger.error(
            f"Patient with ID {self.patient_id} has no series available that correlates with the image '{series_key}'. "
            f"The expected values of specified tag for this image are {self.tag_values[series_key]} while the "
            f"patient record only contains the following values: {self.available_tag_values}."
        )

        self.failed_images.append(series_key)

    def get_patient_dataset(
            self,
            transforms: Optional[Union[Compose, DataTransform, MonaiMapTransform, PhysicalSpaceTransform]] = None
    ) -> PatientDataModel:
        """
        Get the patient dataset.

        Parameters
        ----------
        transforms : Union[Compose, DataTransform, MonaiMapTransform, PhysicalSpaceTransform]
            A sequence of transformations to apply. PhysicalSpaceTransform are applied in the physical space, i.e on
            the SimpleITK image, while MonaiMapTransform are applied in the array space, i.e on the numpy array that
            represents the image. DataTransform transforms the data using other a patient's other images or
            segmentations. The keys for images are assumed to be the arbitrary series key set in 'tag_values'.
            For segmentation, keys are organ names. Note that if 'tag_values' is None, the keys for images are
            assumed to be modalities.

        Returns
        -------
        patient_dataset : PatientDataModel
            A named tuple grouping the patient's data extracted from its DICOM files and the patient's medical image
            segmentation data extracted from the segmentation files.
        """
        patient_data_context = PatientDataQueryContext(
            path_to_patient_folder=self._path_to_patient_folder,
            paths_to_segmentations=self.paths_to_segmentations,
            tag_values=self._tag_values,
            tag=self.tag,
            organs=self._organs,
            erase_unused_dicom_files=self._erase_unused_dicom_files
        )
        patient_dataset = patient_data_context.create_patient_data()

        if transforms:
            apply_transforms(patient_dataset=patient_dataset, transforms=transforms)

        _logger.debug(
            f"Chosen patient data query strategy: '{patient_data_context.patient_data_query_strategy.name}'."
        )
        _logger.info(f"{len(patient_dataset.data)} images added to the patient dataset, namely: ")

        for image_and_segmentation_data in patient_dataset.data:
            image = image_and_segmentation_data.image
            segmentations = image_and_segmentation_data.segmentations

            segmented_organs = set()
            if segmentations:
                for segmentation in segmentations:
                    for organ in list(segmentation.simple_itk_label_maps.keys()):
                        segmented_organs.add(organ)

            if hasattr(image.dicom_header, "SeriesDescription"):
                series_description = image.dicom_header.SeriesDescription
                _logger.info(f"Series Description : {series_description}")

            if hasattr(image.dicom_header, "Modality"):
                modality = image.dicom_header.Modality
                _logger.info(f"  Modality : {modality}")

            image_segmentation_available = True if segmentations else False
            _logger.info(f"  Image Segmentation available: {image_segmentation_available}")

            if image_segmentation_available:
                _logger.info(f"  Segmented organs : {segmented_organs}")

        patient_dataset.transforms_history = TransformsHistory(transforms)

        return patient_dataset
